@@ -1,43 +1,27 @@
 """scrapli.driver.network_driver"""
-import logging
-import re
-import warnings
-from abc import ABC, abstractmethod
-from collections import namedtuple
-from functools import lru_cache
-from typing import Any, Dict, List, Optional, Tuple, Union
+from collections import UserList
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
 
+from scrapli.driver.base_network_driver import NetworkDriverBase, PrivilegeAction, PrivilegeLevel
 from scrapli.driver.generic_driver import GenericDriver
-from scrapli.exceptions import UnknownPrivLevel
-from scrapli.helper import get_prompt_pattern, resolve_file
-from scrapli.response import Response
+from scrapli.exceptions import CouldNotAcquirePrivLevel
+from scrapli.response import MultiResponse, Response
 
-PrivilegeLevel = namedtuple(
-    "PrivilegeLevel",
-    "pattern "  # regex pattern for this priv level
-    "name "  # friendly name of the priv level
-    "previous_priv "  # name of the "lower"/"previous" priv level
-    "deescalate "  # how to deescalate *from* this priv level (to the lower/previous priv)
-    "escalate "  # how to escalate *to* this priv level (from the lower/previous priv)
-    "escalate_auth "  # is there auth to get *to* this priv level
-    "escalate_prompt ",  # what is the prompt to get *to* this priv level
-)
-
-NoPrivLevel = PrivilegeLevel("", "", "", "", "", "", "")
+if TYPE_CHECKING:
+    ScrapliMultiResponse = UserList[Response]  # pylint:  disable=E1136; # pragma:  no cover
+else:
+    ScrapliMultiResponse = UserList
 
 
-PRIVS: Dict[str, PrivilegeLevel] = {}
-
-LOG = logging.getLogger("driver")
-
-
-class NetworkDriver(GenericDriver, ABC):
-    @abstractmethod
+class NetworkDriver(GenericDriver, NetworkDriverBase):
     def __init__(
         self,
         privilege_levels: Dict[str, PrivilegeLevel],
         default_desired_privilege_level: str,
         auth_secondary: str = "",
+        failed_when_contains: Optional[List[str]] = None,
+        textfsm_platform: str = "",
+        genie_platform: str = "",
         **kwargs: Any,
     ):
         """
@@ -50,6 +34,9 @@ class NetworkDriver(GenericDriver, ABC):
                 like that upon first login, and is also the priv level scrapli will try to acquire
                 for normal "command" operations (`send_command`, `send_commands`)
             auth_secondary: password to use for secondary authentication (enable)
+            failed_when_contains: list of strings that indicate a command/configuration has failed
+            textfsm_platform: string name of platform to use for textfsm parsing
+            genie_platform: string name of platform to use for genie parsing
             **kwargs: keyword args to pass to inherited class(es)
 
         Returns:
@@ -59,79 +46,19 @@ class NetworkDriver(GenericDriver, ABC):
             N/A
 
         """
-        self.privilege_levels = privilege_levels
-        self.default_desired_privilege_level = default_desired_privilege_level
-
-        self.comms_prompt_pattern_all = r"|".join(
-            rf"(?P<{priv_level}>{priv_level_data.pattern})"
-            for priv_level, priv_level_data in self.privilege_levels.items()
-        )
-        self._current_priv_level = NoPrivLevel
-        self._priv_map: Dict[str, List[str]] = {}
-        self._build_priv_map()
+        self._check_kwargs_comms_prompt_pattern(kwargs=kwargs)
 
         self.auth_secondary = auth_secondary
+        self.privilege_levels = privilege_levels
+        self._priv_map = {}
+        self.default_desired_privilege_level = default_desired_privilege_level
+        self.update_privilege_levels(update_channel=False)
 
-        self.textfsm_platform: str = ""
-        self.genie_platform: str = ""
-        self.failed_when_contains: List[str] = []
+        self.textfsm_platform = textfsm_platform
+        self.genie_platform = genie_platform
+        self.failed_when_contains = failed_when_contains or []
 
-        super().__init__(comms_prompt_pattern=self.comms_prompt_pattern_all, **kwargs)
-
-    def _build_priv_map(self) -> None:
-        """
-        Build a "map" of privilege levels
-
-        `_priv_map` is a "map" of all privilege levels mapped out to the lowest available priv. This
-        map is used for determining how to escalate/deescalate.
-
-        Args:
-            N/A
-
-        Returns:
-            N/A  # noqa: DAR202
-
-        Raises:
-            N/A
-
-        """
-        for priv_level in self.privilege_levels:
-            self._priv_map[priv_level] = [priv_level]
-            while True:
-                previous_priv = self.privilege_levels[self._priv_map[priv_level][0]].previous_priv
-                if not previous_priv:
-                    break
-                self._priv_map[priv_level].insert(0, previous_priv)
-
-    @lru_cache()
-    def _determine_current_priv(self, current_prompt: str) -> PrivilegeLevel:
-        """
-        Determine current privilege level from prompt string
-
-        Args:
-            current_prompt: string of current prompt
-
-        Returns:
-            PrivilegeLevel: NamedTuple of current privilege level
-
-        Raises:
-            UnknownPrivLevel: if privilege level cannot be determined
-
-        """
-        prompt_pattern = get_prompt_pattern(prompt="", class_prompt=self.comms_prompt_pattern_all)
-        search_result = re.search(pattern=prompt_pattern, string=current_prompt.encode())
-        if search_result:
-            possible_priv_levels = [
-                priv_name
-                for priv_name, matched_prompt in search_result.groupdict().items()
-                if matched_prompt
-            ]
-            priv_level = self.privilege_levels[possible_priv_levels[0]]
-            LOG.debug(f"Determined current privilege level is `{priv_level.name}`")
-            return priv_level
-        raise UnknownPrivLevel(
-            f"Could not determine privilege level from provided prompt: `{current_prompt}`"
-        )
+        super().__init__(comms_prompt_pattern=self.comms_prompt_pattern, **kwargs)
 
     def _escalate(self, escalate_priv: PrivilegeLevel) -> None:
         """
@@ -147,35 +74,17 @@ class NetworkDriver(GenericDriver, ABC):
             N/A
 
         """
-        next_prompt_pattern = escalate_priv.pattern
+        self._pre_escalate(escalate_priv=escalate_priv)
+
         if escalate_priv.escalate_auth is True:
-            if not self.auth_secondary:
-                err = (
-                    "Privilege escalation generally requires an `auth_secondary` password, "
-                    "but none is set!"
-                )
-                msg = f"***** {err} {'*' * (80 - len(err))}"
-                fix = (
-                    "scrapli will try to escalate privilege without entering a password but may "
-                    "fail.\nSet an `auth_secondary` password if your device requires a password to "
-                    "increase privilege, otherwise ignore this message."
-                )
-                warning = "\n" + msg + "\n" + fix + "\n" + msg
-                warnings.warn(warning)
-            else:
-                escalate_cmd: str = escalate_priv.escalate
-                escalate_prompt: str = escalate_priv.escalate_prompt
-                escalate_auth = self.auth_secondary
-                super().send_interactive(
-                    interact_events=[
-                        (escalate_cmd, escalate_prompt, False),
-                        (escalate_auth, next_prompt_pattern, True),
-                    ],
-                )
-                self.channel.comms_prompt_pattern = next_prompt_pattern
-                return
-        self.channel.comms_prompt_pattern = next_prompt_pattern
-        self.channel.send_input(channel_input=escalate_priv.escalate)
+            super().send_interactive(
+                interact_events=[
+                    (escalate_priv.escalate, escalate_priv.escalate_prompt, False),
+                    (self.auth_secondary, escalate_priv.pattern, True),
+                ],
+            )
+        else:
+            self.channel.send_input(channel_input=escalate_priv.escalate)
 
     def _deescalate(self, current_priv: PrivilegeLevel) -> None:
         """
@@ -191,9 +100,6 @@ class NetworkDriver(GenericDriver, ABC):
             N/A
 
         """
-        self.channel.comms_prompt_pattern = self.privilege_levels[
-            current_priv.previous_priv
-        ].pattern
         self.channel.send_input(channel_input=current_priv.deescalate)
 
     def acquire_priv(self, desired_priv: str) -> None:
@@ -201,69 +107,39 @@ class NetworkDriver(GenericDriver, ABC):
         Acquire desired priv level
 
         Args:
-            desired_priv: string name of desired privilege level
-                (see scrapli.driver.<driver_category.device_type>.driver for levels)
+            desired_priv: string name of desired privilege level see
+                `scrapli.driver.<driver_category.device_type>.driver` for levels
 
         Returns:
             N/A  # noqa: DAR202
 
         Raises:
-           UnknownPrivLevel: if attempting to acquire an unknown priv
+            CouldNotAcquirePrivLevel: if desired_priv cannot be attained
 
         """
-        LOG.info(f"Attempting to acquire `{desired_priv}` privilege level")
-        if desired_priv not in self.privilege_levels.keys():
-            raise UnknownPrivLevel(
-                f"Requested privilege level `{desired_priv}` not a valid privilege level of "
-                f"`{self.__class__.__name__}`"
-            )
-
-        map_to_desired_priv = self._priv_map[desired_priv]
+        resolved_priv, map_to_desired_priv = self._pre_acquire_priv(desired_priv=desired_priv)
 
+        privilege_change_count = 0
         while True:
-            # if we are already at the desired priv, we don't need to do any thing
-            current_priv = self._determine_current_priv(current_prompt=self.channel.get_prompt())
-            if current_priv.name == desired_priv:
-                self._current_priv_level = current_priv
-                return
-
-            map_to_current_priv = self._priv_map[current_priv.name]
-            priv_map = (
-                map_to_current_priv
-                if map_to_current_priv > map_to_desired_priv
-                else map_to_desired_priv
+            current_prompt = self.channel.get_prompt()
+            privilege_action, target_priv = self._process_acquire_priv(
+                resolved_priv=resolved_priv,
+                map_to_desired_priv=map_to_desired_priv,
+                current_prompt=current_prompt,
             )
 
-            desired_priv_index = priv_map.index(desired_priv)
-            current_priv_index = priv_map.index(current_priv.name)
-            if current_priv_index > desired_priv_index:
-                deescalate_priv = priv_map[current_priv_index - 1]
-                LOG.info(f"Attempting to deescalate from {current_priv.name} to {deescalate_priv}")
-                self._deescalate(current_priv=current_priv)
-            else:
-                escalate_priv = self.privilege_levels[priv_map[current_priv_index + 1]]
-                LOG.info(f"Attempting to escalate from {current_priv.name} to {escalate_priv.name}")
-                self._escalate(escalate_priv=escalate_priv)
+            if privilege_action == PrivilegeAction.NO_ACTION:
+                self._current_priv_level = target_priv
+                return
+            if privilege_action == PrivilegeAction.DEESCALATE:
+                self._deescalate(current_priv=target_priv)
+            if privilege_action == PrivilegeAction.ESCALATE:
+                self._escalate(escalate_priv=target_priv)
 
-    def _update_response(self, response: Response) -> None:
-        """
-        Update response with network driver specific data
-
-        This happens here as the underlying channel provides a response object but is unaware of any
-        of the network/platform specific attributes that may need to get updated
-
-        Args:
-            response: response to update
-
-        Returns:
-            N/A  # noqa: DAR202
-
-        Raises:
-            N/A
-
-        """
-        response.textfsm_platform = self.textfsm_platform
-        response.genie_platform = self.genie_platform
+            privilege_change_count += 1
+            if privilege_change_count > len(self.privilege_levels) * 2:
+                msg = f"Failed to acquire requested privilege level {resolved_priv}"
+                raise CouldNotAcquirePrivLevel(msg)
 
     def send_command(
         self,
@@ -308,7 +184,7 @@ class NetworkDriver(GenericDriver, ABC):
         strip_prompt: bool = True,
         failed_when_contains: Optional[Union[str, List[str]]] = None,
         stop_on_failed: bool = False,
-    ) -> List[Response]:
+    ) -> ScrapliMultiResponse:
         """
         Send multiple commands
 
@@ -322,7 +198,7 @@ class NetworkDriver(GenericDriver, ABC):
                 as of current execution
 
         Returns:
-            responses: list of Scrapli Response objects
+            ScrapliMultiResponse: Scrapli MultiResponse object
 
         Raises:
             N/A
@@ -352,7 +228,7 @@ class NetworkDriver(GenericDriver, ABC):
         strip_prompt: bool = True,
         failed_when_contains: Optional[Union[str, List[str]]] = None,
         stop_on_failed: bool = False,
-    ) -> List[Response]:
+    ) -> ScrapliMultiResponse:
         """
         Send command(s) from file
 
@@ -364,20 +240,13 @@ class NetworkDriver(GenericDriver, ABC):
                 as of current execution
 
         Returns:
-            responses: list of Scrapli Response objects
+            ScrapliMultiResponse: Scrapli MultiResponse object
 
         Raises:
-            TypeError: if anything but a string is provided for `file`
+            N/A
 
         """
-        if not isinstance(file, str):
-            raise TypeError(
-                f"`send_commands_from_file` expects a string path to file, got {type(file)}"
-            )
-        resolved_file = resolve_file(file)
-
-        with open(resolved_file, "r") as f:
-            commands = f.read().splitlines()
+        commands = self._pre_send_commands_from_file(file=file)
 
         return self.send_commands(
             commands=commands,
@@ -390,6 +259,7 @@ class NetworkDriver(GenericDriver, ABC):
         self,
         interact_events: List[Tuple[str, str, Optional[bool]]],
         failed_when_contains: Optional[Union[str, List[str]]] = None,
+        privilege_level: str = "",
     ) -> Response:
         """
         Interact with a device with changing prompts per input.
@@ -444,6 +314,7 @@ class NetworkDriver(GenericDriver, ABC):
                 not provided it is assumed the input is "normal" (not hidden)
             failed_when_contains: list of strings that, if present in final output, represent a
                 failed command/interaction
+            privilege_level: name of the privilege level to operate in
 
         Returns:
             Response: scrapli Response object
@@ -452,8 +323,15 @@ class NetworkDriver(GenericDriver, ABC):
             N/A
 
         """
-        if self._current_priv_level.name != self.default_desired_privilege_level:
-            self.acquire_priv(desired_priv=self.default_desired_privilege_level)
+        if privilege_level:
+            resolved_privilege_level = self._get_privilege_level_name(
+                requested_priv=privilege_level
+            )
+        else:
+            resolved_privilege_level = self.default_desired_privilege_level
+
+        if self._current_priv_level.name != resolved_privilege_level:
+            self.acquire_priv(desired_priv=resolved_privilege_level)
         response = super().send_interactive(
             interact_events=interact_events, failed_when_contains=failed_when_contains
         )
@@ -463,7 +341,7 @@ class NetworkDriver(GenericDriver, ABC):
 
     def _abort_config(self) -> None:
         """
-        Abort a configuration session for platforms with commit type functionality; i.e iosxr/junos
+        Abort a configuration operation/session if applicable (for config sessions like junos/iosxr)
 
         Args:
             N/A
@@ -476,40 +354,92 @@ class NetworkDriver(GenericDriver, ABC):
 
         """
 
-    def send_configs(
+    def send_config(
         self,
-        configs: Union[str, List[str]],
+        config: str,
         strip_prompt: bool = True,
         failed_when_contains: Optional[Union[str, List[str]]] = None,
         stop_on_failed: bool = False,
-    ) -> List[Response]:
+        privilege_level: str = "",
+    ) -> Response:
         """
         Send configuration(s)
 
         Args:
-            configs: string or list of strings to send to device in config mode
+            config: string configuration to send to the device, supports sending multi-line strings
             strip_prompt: True/False strip prompt from returned output
             failed_when_contains: string or list of strings indicating failure if found in response
             stop_on_failed: True/False stop executing commands if a command fails, returns results
-                as of current execution
+                as of current execution; aborts configuration session if applicable (iosxr/junos or
+                eos/nxos if using a configuration session)
+            privilege_level: name of configuration privilege level/type to acquire; this is platform
+                dependent, so check the device driver for specifics. Examples of privilege_name
+                would be "configuration_exclusive" for IOSXRDriver, or "configuration_private" for
+                JunosDriver. You can also pass in a name of a configuration session such as
+                "my-config-session" if you have registered a session using the
+                "register_config_session" method of the EOSDriver or NXOSDriver.
 
         Returns:
-            responses: List of Scrape Response objects
+            Response: Scrapli Response object
 
         Raises:
             N/A
 
         """
-        if isinstance(configs, str):
-            configs = [configs]
+        split_config = self._pre_send_config(config=config)
 
-        if not self._current_priv_level.name.startswith("configuration"):
-            self.acquire_priv(desired_priv="configuration")
+        # now that we have a list of configs, just use send_configs to actually execute them
+        multi_response = self.send_configs(
+            configs=split_config,
+            strip_prompt=strip_prompt,
+            failed_when_contains=failed_when_contains,
+            stop_on_failed=stop_on_failed,
+            privilege_level=privilege_level,
+        )
+        return self._post_send_config(config=config, multi_response=multi_response)
 
-        if failed_when_contains is None:
-            failed_when_contains = self.failed_when_contains
+    def send_configs(
+        self,
+        configs: List[str],
+        strip_prompt: bool = True,
+        failed_when_contains: Optional[Union[str, List[str]]] = None,
+        stop_on_failed: bool = False,
+        privilege_level: str = "",
+    ) -> ScrapliMultiResponse:
+        """
+        Send configuration(s)
 
-        responses = []
+        Args:
+            configs: list of strings to send to device in config mode
+            strip_prompt: True/False strip prompt from returned output
+            failed_when_contains: string or list of strings indicating failure if found in response
+            stop_on_failed: True/False stop executing commands if a command fails, returns results
+                as of current execution; aborts configuration session if applicable (iosxr/junos or
+                eos/nxos if using a configuration session)
+            privilege_level: name of configuration privilege level/type to acquire; this is platform
+                dependent, so check the device driver for specifics. Examples of privilege_name
+                would be "configuration_exclusive" for IOSXRDriver, or "configuration_private" for
+                JunosDriver. You can also pass in a name of a configuration session such as
+                "my-config-session" if you have registered a session using the
+                "register_config_session" method of the EOSDriver or NXOSDriver.
+
+        Returns:
+            ScrapliMultiResponse: Scrapli MultiResponse object
+
+        Raises:
+            N/A
+
+        """
+        resolved_privilege_level, failed_when_contains = self._pre_send_configs(
+            configs=configs,
+            failed_when_contains=failed_when_contains,
+            privilege_level=privilege_level,
+        )
+
+        if self._current_priv_level.name != resolved_privilege_level:
+            self.acquire_priv(desired_priv=resolved_privilege_level)
+
+        responses = MultiResponse()
         _failed_during_execution = False
         for config in configs:
             response = super().send_command(
@@ -518,18 +448,15 @@ class NetworkDriver(GenericDriver, ABC):
                 failed_when_contains=failed_when_contains,
             )
             responses.append(response)
-            if stop_on_failed is True and response.failed is True:
+            if response.failed is True:
                 _failed_during_execution = True
-                break
-
-        for response in responses:
-            self._update_response(response=response)
+                if stop_on_failed is True:
+                    break
 
         if _failed_during_execution is True:
             self._abort_config()
 
-        self.acquire_priv(desired_priv=self.default_desired_privilege_level)
-        return responses
+        return self._post_send_configs(responses=responses)
 
     def send_configs_from_file(
         self,
@@ -537,7 +464,8 @@ class NetworkDriver(GenericDriver, ABC):
         strip_prompt: bool = True,
         failed_when_contains: Optional[Union[str, List[str]]] = None,
         stop_on_failed: bool = False,
-    ) -> List[Response]:
+        privilege_level: str = "",
+    ) -> ScrapliMultiResponse:
         """
         Send configuration(s) from a file
 
@@ -546,27 +474,28 @@ class NetworkDriver(GenericDriver, ABC):
             strip_prompt: True/False strip prompt from returned output
             failed_when_contains: string or list of strings indicating failure if found in response
             stop_on_failed: True/False stop executing commands if a command fails, returns results
-                as of current execution
+                as of current execution; aborts configuration session if applicable (iosxr/junos or
+                eos/nxos if using a configuration session)
+            privilege_level: name of configuration privilege level/type to acquire; this is platform
+                dependent, so check the device driver for specifics. Examples of privilege_name
+                would be "exclusive" for IOSXRDriver, "private" for JunosDriver. You can also pass
+                in a name of a configuration session such as "session_mysession" if you have
+                registered a session using the "register_config_session" method of the EOSDriver or
+                NXOSDriver.
 
         Returns:
-            responses: List of Scrape Response objects
+            ScrapliMultiResponse: Scrapli MultiResponse object
 
         Raises:
-            TypeError: if anything but a string is provided for `file`
+            N/A
 
         """
-        if not isinstance(file, str):
-            raise TypeError(
-                f"`send_configs_from_file` expects a string path to file, got {type(file)}"
-            )
-        resolved_file = resolve_file(file)
-
-        with open(resolved_file, "r") as f:
-            configs = f.read().splitlines()
+        configs = self._pre_send_configs_from_file(file=file)
 
         return self.send_configs(
             configs=configs,
             strip_prompt=strip_prompt,
             failed_when_contains=failed_when_contains,
             stop_on_failed=stop_on_failed,
+            privilege_level=privilege_level,
         )
