@@ -5,6 +5,10 @@ class ScrapliException(Exception):
     """Base Exception for scrapli"""
 
 
+class ScrapliCommandFailure(ScrapliException):
+    """Exception for scrapli command/config failures"""
+
+
 class ScrapliTimeout(ScrapliException):
     """Exception for any scrapli timeouts"""
 
@@ -35,3 +39,11 @@ class CouldNotAcquirePrivLevel(ScrapliException):
 
 class TransportPluginError(ScrapliException):
     """Exception for transport plugin loading errors"""
+
+
+class ConnectionNotOpened(ScrapliException):
+    """Exception for failures due to connection not being open"""
+
+
+class UnsupportedPlatform(ScrapliException):
+    """Exception for unsupported platform; i.e. using system transport on windows"""
