@@ -10,12 +10,9 @@ logging.basicConfig(
 )
 logger = logging.getLogger("scrapli")
 
-conn = EOSDriver(**eos_device)
+conn = EOSDriver(**eos_device, auth_secondary="VR-netlab9")
 conn.open()
 
-# vrnetlab eos has no enable password so we just enable this manually for now
-# conn.channel.send_inputs("enable")
-
 print("***** Get Prompt:")
 prompt = conn.get_prompt()
 print(prompt)
